@@ -11,7 +11,6 @@ use std::f64;
 use std::fmt;
 use std::iter;
 use std::marker::PhantomData;
-use std::mem::discriminant;
 use std::str;
 use std::vec;
 
@@ -147,10 +146,6 @@ enum ErrorKind {
         found: &'static str,
     },
 
-    /// An array was decoded but the types inside of it were mixed, which is
-    /// disallowed by TOML.
-    MixedArrayType,
-
     /// A duplicate table definition was found.
     DuplicateTable(String),
 
@@ -196,6 +191,9 @@ enum ErrorKind {
         available: &'static [&'static str],
     },
 
+    /// Unquoted string was found when quoted one was expected
+    UnquotedString,
+
     #[doc(hidden)]
     __Nonexhaustive,
 }
@@ -338,7 +336,7 @@ fn build_table_indices<'de>(tables: &[Table<'de>]) -> HashMap<Vec<Cow<'de, str>>
     let mut res = HashMap::new();
     for (i, table) in tables.iter().enumerate() {
         let header = table.header.iter().map(|v| v.1.clone()).collect::<Vec<_>>();
-        res.entry(header).or_insert(Vec::new()).push(i);
+        res.entry(header).or_insert_with(Vec::new).push(i);
     }
     res
 }
@@ -364,7 +362,7 @@ fn build_table_pindices<'de>(tables: &[Table<'de>]) -> HashMap<Vec<Cow<'de, str>
         let header = table.header.iter().map(|v| v.1.clone()).collect::<Vec<_>>();
         for len in 0..=header.len() {
             res.entry(header[..len].to_owned())
-                .or_insert(Vec::new())
+                .or_insert_with(Vec::new)
                 .push(i);
         }
     }
@@ -685,7 +683,7 @@ impl<'de, 'b> de::Deserializer<'de> for MapVisitor<'de, 'b> {
         }
         let table = &mut self.tables[0];
         let values = table.values.take().expect("table has no values?");
-        if table.header.len() == 0 {
+        if table.header.is_empty() {
             return Err(self.de.error(self.cur, ErrorKind::EmptyTableKey));
         }
         let name = table.header[table.header.len() - 1].1.to_owned();
@@ -1433,7 +1431,7 @@ impl<'a> Deserializer<'a> {
                 start,
                 end,
             },
-            Some((span, Token::Keylike(key))) => self.number_or_date(span, key)?,
+            Some((span, Token::Keylike(key))) => self.parse_keylike(at, span, key)?,
             Some((span, Token::Plus)) => self.number_leading_plus(span)?,
             Some((Span { start, .. }, Token::LeftBrace)) => {
                 self.inline_table().map(|(Span { end, .. }, table)| Value {
@@ -1456,13 +1454,25 @@ impl<'a> Deserializer<'a> {
                         expected: "a value",
                         found: token.1.describe(),
                     },
-                ))
+                ));
             }
             None => return Err(self.eof()),
         };
         Ok(value)
     }
 
+    fn parse_keylike(&mut self, at: usize, span: Span, key: &'a str) -> Result<Value<'a>, Error> {
+        if key == "inf" || key == "nan" {
+            return self.number_or_date(span, key);
+        }
+
+        let first_char = key.chars().next().expect("key should not be empty here");
+        match first_char {
+            '-' | '0'..='9' => self.number_or_date(span, key),
+            _ => Err(self.error(at, ErrorKind::UnquotedString)),
+        }
+    }
+
     fn number_or_date(&mut self, span: Span, s: &'a str) -> Result<Value<'a>, Error> {
         if s.contains('T')
             || s.contains('t')
@@ -1638,7 +1648,7 @@ impl<'a> Deserializer<'a> {
 
             if c == '0' && first {
                 first_zero = true;
-            } else if c.to_digit(radix).is_some() {
+            } else if c.is_digit(radix) {
                 if !first && first_zero && !allow_leading_zeros {
                     return Err(self.error(at, ErrorKind::NumberInvalid));
                 }
@@ -1827,13 +1837,7 @@ impl<'a> Deserializer<'a> {
             if let Some(span) = self.eat_spanned(Token::RightBracket)? {
                 return Ok((span, ret));
             }
-            let at = self.tokens.current();
             let value = self.value()?;
-            if let Some(last) = ret.last() {
-                if !value.same_type(last) {
-                    return Err(self.error(at, ErrorKind::MixedArrayType));
-                }
-            }
             ret.push(value);
             intermediate(self)?;
             if !self.eat(Token::Comma)? {
@@ -2087,7 +2091,7 @@ impl std::convert::From<Error> for std::io::Error {
 
 impl fmt::Display for Error {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-        match self.inner.kind {
+        match &self.inner.kind {
             ErrorKind::UnexpectedEof => "unexpected eof encountered".fmt(f)?,
             ErrorKind::InvalidCharInString(c) => write!(
                 f,
@@ -2118,7 +2122,6 @@ impl fmt::Display for Error {
             }
             ErrorKind::NumberInvalid => "invalid number".fmt(f)?,
             ErrorKind::DateInvalid => "invalid date".fmt(f)?,
-            ErrorKind::MixedArrayType => "mixed types in an array".fmt(f)?,
             ErrorKind::DuplicateTable(ref s) => {
                 write!(f, "redefinition of table `{}`", s)?;
             }
@@ -2143,6 +2146,10 @@ impl fmt::Display for Error {
                 "unexpected keys in table: `{:?}`, available keys: `{:?}`",
                 keys, available
             )?,
+            ErrorKind::UnquotedString => write!(
+                f,
+                "invalid TOML value, did you mean to use a quoted string?"
+            )?,
             ErrorKind::__Nonexhaustive => panic!(),
         }
 
@@ -2165,36 +2172,7 @@ impl fmt::Display for Error {
     }
 }
 
-impl error::Error for Error {
-    fn description(&self) -> &str {
-        match self.inner.kind {
-            ErrorKind::UnexpectedEof => "unexpected eof encountered",
-            ErrorKind::InvalidCharInString(_) => "invalid char in string",
-            ErrorKind::InvalidEscape(_) => "invalid escape in string",
-            ErrorKind::InvalidHexEscape(_) => "invalid hex escape in string",
-            ErrorKind::InvalidEscapeValue(_) => "invalid escape value in string",
-            ErrorKind::NewlineInString => "newline in string found",
-            ErrorKind::Unexpected(_) => "unexpected or invalid character",
-            ErrorKind::UnterminatedString => "unterminated string",
-            ErrorKind::NewlineInTableKey => "found newline in table key",
-            ErrorKind::Wanted { .. } => "expected a token but found another",
-            ErrorKind::NumberInvalid => "invalid number",
-            ErrorKind::DateInvalid => "invalid date",
-            ErrorKind::MixedArrayType => "mixed types in an array",
-            ErrorKind::DuplicateTable(_) => "duplicate table",
-            ErrorKind::RedefineAsArray => "table redefined as array",
-            ErrorKind::EmptyTableKey => "empty table key found",
-            ErrorKind::MultilineStringKey => "invalid multiline string for key",
-            ErrorKind::Custom => "a custom error",
-            ErrorKind::ExpectedTuple(_) => "expected table length",
-            ErrorKind::ExpectedTupleIndex { .. } => "expected table key",
-            ErrorKind::ExpectedEmptyTable => "expected empty table",
-            ErrorKind::DottedKeyInvalidType => "dotted key invalid type",
-            ErrorKind::UnexpectedKeys { .. } => "unexpected keys in table",
-            ErrorKind::__Nonexhaustive => panic!(),
-        }
-    }
-}
+impl error::Error for Error {}
 
 impl de::Error for Error {
     fn custom<T: fmt::Display>(msg: T) -> Error {
@@ -2283,9 +2261,3 @@ impl<'a> E<'a> {
         }
     }
 }
-
-impl<'a> Value<'a> {
-    fn same_type(&self, other: &Value<'a>) -> bool {
-        discriminant(&self.e) == discriminant(&other.e)
-    }
-}
